
#ifndef HTCONFIG_H
#define HTCONFIG_H

#include "HTUtils.h"
#include "HTList.h"
#include "HTAccess.h"


#define DEFAULT_INPUT_TIMEOUT	 120	/*  120 =  2 min */
#define DEFAULT_OUTPUT_TIMEOUT	1200	/* 1200 = 20 min */
#define DEFAULT_SCRIPT_TIMEOUT	 600	/*  600 = 10 min */
#define DEFAULT_CLEANUP_TIMEOUT   15	/* 15 secs to write to log file	*/

extern char * HTBinDir;			/* Physical /htbin directory	*/
                                        /* path (this is obsolite, and	*/
                                        /* only used if HTSearchScript	*/
                                        /* is given as non-absolute --	*/
                                        /* and this is only for rule	*/
                                        /* file backward-compatibility.	*/
extern BOOL   HTImProxy;		/* Am I a proxy server		*/
extern char * HTSearchScript;		/* Search script name.		*/
extern char * HTPutScript;		/* Script to handle PUT		*/
extern char * HTPostScript;		/* Script to handle POST	*/
extern char * HTServerHostName;		/* Used for CGI scripts		*/
extern char * HTPortAddr;		/* Default port address string	*/
extern char * HTProxyCacheRoot;		/* Cache directory for gateway	*/
extern char * HTMetaDir;		/* Directory for metafiles	*/
extern char * HTMetaSuffix;		/* Suffix for metafiles		*/
extern HTList * HTNoLog;		/* Machines for which no logging*/
extern HTList * HTNoCaching;		/* No caching if matching these	*/
extern BOOL HTDisabled[(int)MAX_METHODS + 1];	/* Disabled methods	*/
extern BOOL HTDoSetUid;			/* Do setuid() to auth'd uid	*/

extern char * HTCacheAccessLog;		/* Cache access log		*/
extern BOOL HTNewLogFileFormat;		/* Use new common logfile format*/
extern BOOL HTUseGMT;			/* Use GMT instead of localtime	*/
extern BOOL HTNoBackGround;		/* Don't auto-go background	*/
extern BOOL HTStandAlone;		/* Am I standalone?		*/
extern char * HTUserDir;		/* User supported directory name*/
extern char * HTUserId;			/* Default user id		*/
extern char * HTGroupId;		/* Default group id		*/

extern BOOL    always_welcome;		/* Redirect directory names to	*/
                                        /* welcome page on that dir.	*/

extern BOOL    do_rfc931;		/* Should we check remote ident	*/

extern BOOL do_accept_hack;
        /* Do we send the document even if there was no corresponding	*/
        /* Accept header line. This is because even currently browsers	*/
        /* which are NOT sending Accept's are widely used. I'll disable	*/
        /* this feature by default soon. For now it makes things easier.*/

extern char * server_root;		/* Server's "home directory"	*/
extern BOOL icons_inited;		/* If not we'll use defaults	*/

extern int HTInputTimeOut;		/* Timeout for request		*/
extern int HTOutputTimeOut;		/* Timeout for response		*/
extern int HTScriptTimeOut;		/* Timeout for scripts		*/

extern HTList* cache_clean;		/* Forced expiry according templ*/
extern time_t  cache_clean_def;		/* Remove if older than 3 weeks	*/
extern HTList* cache_exp;		/* Expiry according to template	*/
extern time_t  cache_exp_def;		/* Default expiry to use	*/
extern HTList* cache_unused;		/* Removal according to template*/
extern time_t  cache_unused_def;	/* Remove if not used in 2 weeks*/
extern long    cache_max_k;		/* Default size 1MB (small)	*/
extern int     cache_max_f;		/* At most this many cache files*/
extern int     cache_limit_1;		/* Size no problem until 200K	*/
extern int     cache_limit_2;		/* No caching after 4M		*/
extern float   cache_lm_factor;		/* For approximating Expires:	*/
extern time_t  cache_time_margin;	/* Time accuracy		*/
extern BOOL    cache_exp_check;		/* Do expiry check		*/

extern BOOL    cache_no_connect;
	/* Use only local cache and return error if the document was	*/
	/* not in the cache.						*/

extern BOOL    gc_disabled;		/* Never do garbage collection	*/
extern int     gc_req_interval;		/* gc every 10000 reqs		*/
extern time_t  gc_time_interval;	/* gc once a day		*/
extern int     gc_mem_usage_control;	/* How radical memory usage	*/
extern time_t  cache_lock_timeout;	/* Timeout for cache lockfiles	*/
extern time_t  gc_daily_gc;		/* daily gc (default at 3:00am)	*/

extern char *  pid_file;		/* File to write the pid number	*/

PUBLIC time_t get_default_expiry PARAMS((char * url));
PUBLIC time_t get_unused PARAMS((char * url));
PUBLIC time_t get_clean PARAMS((char * url));
PUBLIC void HTPrependPass PARAMS((CONST char *	pattern,
				  CONST char *	equiv));

#endif /* HTCONFIG_H */

