
/* MODULE						HTRequest.c
**		HTTP REQUEST PARSE AND TRANSLATION
**
** AUTHORS:
**	AL	Ari Luotonen    luotonen@dxcern.cern.ch
**	MD	Mark Donszelmann    duns@vxdeop.cern.ch
**
** HISTORY:
**	11 Dec 93  AL	Written based on the old HTHandle().
**
*/

#include <string.h>
#include <stdio.h>

#include "HTUtils.h"
#include "HTAccess.h"	/* HTRequest */
#include "HTParse.h"	/* HTStrip() */
#include "tcp.h"
#include "HTRequest.h"	/* Implemented here */
#include "HTDaemon.h"
#include "HTConfig.h"
#include "HTTCP.h"	/* HTHostName() */
#include "HTAuth.h"

extern char * HTAppName;
extern char * HTAppVersion;
extern char * HTClientProtocol;
extern char * HTServerProtocol;
extern char * HTClientHost;

PRIVATE HTList * hbuf_head = NULL;
PRIVATE HTList * hbuf_tail = NULL;
PRIVATE int	 hbuf_total_len = 0;

typedef struct _HT_hbuf {
    char *	line;
    BOOL	proxy_header;
    BOOL	cgi_header;
} HT_hbuf;


#define SPACE(c) ((c==' ')||(c=='\t')||(c=='\n')||(c=='\r')) 	/*  DMX */


PRIVATE void hbuf_append ARGS1(char *, line)
{
    if (line && *line) {
	HT_hbuf * node = (HT_hbuf*)calloc(1,sizeof(HT_hbuf));
	if (!node) outofmem(__FILE__, "hbuf_append");

	StrAllocCopy(node->line, line);
	node->proxy_header = YES;
	node->cgi_header = YES;

	if (!hbuf_head) {
	    hbuf_tail = hbuf_head = HTList_new();
	    hbuf_total_len = 0;
	}
	hbuf_tail->next = (HTList*)malloc(sizeof(HTList));
	hbuf_tail = hbuf_tail->next;
	hbuf_tail->object = (void*)node;
	hbuf_tail->next = NULL;
	hbuf_total_len += strlen(line) + 3;
    }
}


PRIVATE void hbuf_proxy_cancel_last NOARGS
{
    if (hbuf_tail) {
	HT_hbuf * node = (HT_hbuf*)(hbuf_tail->object);
	if (node)
	    node->proxy_header = NO;
    }
}


PRIVATE void hbuf_cgi_cancel_last NOARGS
{
    if (hbuf_tail) {
	HT_hbuf * node = (HT_hbuf*)(hbuf_tail->object);
	if (node)
	    node->cgi_header = NO;
    }
}


PRIVATE char * hbuf_proxy_headers ARGS1(HTRequest *, req)
{
    char * ret;
    char * p;
    HT_hbuf * node;
    HTList * cur = hbuf_head;
    int extra = 0;

    if (!hbuf_head || !hbuf_total_len)
	return NULL;

    if (req && HTUserAgent)
	extra = strlen(HTUserAgent);

    p = ret = (char*)malloc(hbuf_total_len + extra + 100);

    while ((node = (HT_hbuf*)HTList_nextObject(cur))) {
	if (node->line && node->proxy_header) {
	    char * line = node->line;
	    while (*line)
		*p++ = *line++;
	    *p++ = CR;
	    *p++ = LF;
	}
    }
    *p = 0;

    sprintf(p, "User-Agent: %s%s%s/%s libwww/%s%c%c",
	    extra		? HTUserAgent		: "",
	    extra		? "  via Gateway  "	: "Gateway ",
	    HTAppName		? HTAppName		: "UnKnown",
	    HTAppVersion	? HTAppVersion		: "0.0",
	    HTLibraryVersion	? HTLibraryVersion	: "0.0",
	    CR, LF);

    return ret;
}


PRIVATE char * do_it_all ARGS4(char *,		env_name,
			       HTList *,	parts,
			       int,		len,
			       char **,		cur)
{
    char * result = NULL;

    if (parts && parts->next) {
	int mylen = 0;
	char * p = (char*)parts->next->object;
	p = strchr(p,':');
	if (p) {
	    p++;
	    while (*p && WHITE(*p)) p++;
	    mylen = strlen(p);
	    len += 2 + mylen;
	}
	result = do_it_all(env_name, parts->next, len, cur);
	if (p) {
	    strcpy(*cur,p);
	    *cur += mylen;
	    if (parts->object) {
		strcpy(*cur, ", ");
		*cur += 2;
	    }
	}
    }
    else if (parts) {
	result = (char*)malloc(len + 20);
	if (!result) outofmem(__FILE__, "do_it_all");

	strcpy(result, env_name);
	*cur = result + strlen(env_name);
    }
    return result;
}



PUBLIC HTList * hbuf_http_env_vars ARGS1(HTRequest *, req)
{
    HTList * result = HTList_new();
    HTList * cur1;

    if (!hbuf_head || !hbuf_head->next)
	return result;

    cur1 = hbuf_head->next;
    while (cur1) {
	HT_hbuf * node1 = (HT_hbuf*)cur1->object;
	char * colon;

	if (node1 && node1->cgi_header && node1->line &&
	    (colon = strchr(node1->line,':'))) {

	    HTList * parts = HTList_new();
	    char * name = NULL;
	    int len;

	    /* get header field name */
	    *colon = 0;
	    StrAllocCopy(name, node1->line);
	    StrAllocCat(name, ":");
	    len = strlen(name);
	    *colon = ':';

	    {	/* collect all the fields that have the same name */
		HTList * cur2 = cur1;
		while (cur2) {
		    HT_hbuf * node2 = (HT_hbuf*)cur2->object;
		    if (node2->cgi_header &&
			!strncasecomp(node2->line,name,len)) {
			node2->cgi_header = NO;
			HTList_addObject(parts, (void*)node2->line);
		    }
		    cur2 = cur2->next;
		}
	    }

	    {	/* make environment variable name */
		char * cur = name;
		char * envname = NULL;

		while (*cur && *cur!=':') {
		    if (*cur=='-') *cur = '_';
		    else *cur = TOUPPER(*cur);
		    cur++;
		}
		if (*cur) *cur = '=';
		StrAllocCopy(envname, "HTTP_");
		StrAllocCat(envname, name);
		free(name);
		name = envname;
	    }

	    {
		/* collapse headers creating a commaseparated list  */
		/* which is an env.var.declaration at the same time */

		char * foo = NULL;
		char * envvar = do_it_all(name, parts, strlen(name), &foo);

		if (HTUser && HTUser->username &&
		    !strncmp(envvar,"HTTP_AUTHORIZATION=",19)) {
		    CTRACE(stderr,
		    "Not......... propagating HTTP_AUTHORIZATION to child\n");
		    FREE(envvar);
		}
		else {
		    HTList_addObject(result, envvar);
		}
		free(name);
		HTList_delete(parts);
	    }
	}
	cur1 = cur1->next;
    }
    return result;
}



PRIVATE void hbuf_free NOARGS
{
    HT_hbuf * node;
    HTList * killme;
    HTList * cur = hbuf_head;

    while (cur) {
	node = (HT_hbuf*)cur->object;
	if (node) {
	    FREE(node->line);
	    free(node);
	}
	killme = cur;
	cur = cur->next;
	free(killme);
    }
    hbuf_head = NULL;
    hbuf_tail = NULL;
    hbuf_total_len = 0;
}


/*				MIME OUTPUT
**				===========
*/

struct _HTStream {
	HTStreamClass *		isa;
	HTStreamClass		class;		/* special per object */
};


PRIVATE BOOL fix_meta_field ARGS1(char *, buf)
{
    char * cur = buf;
    BOOL content = NO;

    if (!buf) return NO;

    while (*cur  &&  *cur!='\r'  &&  *cur!='\n') {
	if (!content  &&  *cur!=' ' &&  *cur!='\t')
	    content = YES;
	cur++;
    }
    if (!content) return NO;

    *cur++ = '\r';
    *cur++ = '\n';
    *cur = 0;
    return YES;
}


PUBLIC char *HTReplyHeaders ARGS1(HTRequest *, req)
{
    return HTReplyHeadersWith(req, NULL, 0);
}


PUBLIC char *HTReplyHeadersWith ARGS3(HTRequest *,	req,
				      char **,		extras,
				      int,		extra_cnt)
{
    char *headers;
    FILE * meta_file = NULL;
    int len;

    if (!HTStatusCode) HTStatusCode = 500;
    if (!HTReasonLine) HTReasonLine = "Internal error";

    len = strlen(HTReasonLine) + 1024 +
	  (HTLocation		? strlen(HTLocation)		: 0) +
	  (HTUri		? strlen(HTUri)			: 0) +
	  (HTMessageId		? strlen(HTMessageId)		: 0) +
	  (HTWWWAuthenticate	? strlen(HTWWWAuthenticate)	: 0);

    if (extras) {
	int i;
	for (i=0; i<extra_cnt; i++) {
	    if (extras[i])
		len += strlen(extras[i]) + 3;
	}
    }

    if (HTMetaFile) {
	struct stat stat_meta;

	if (stat(HTMetaFile, &stat_meta) != -1) {
	    len += stat_meta.st_size;
	    meta_file = fopen(HTMetaFile, "r");
	}
    }

    headers = (char*)malloc(len);
    if (!headers) outofmem(__FILE__, "HTReplyHeaders");

    sprintf(headers,
	    "%s %d %s\r\n%s\r\n%s%s",
	    HTServerProtocol, HTStatusCode, HTReasonLine,
	    "MIME-Version: 1.0",
	    "Server: CERN/", VD);

#ifndef VMS
    if (cur_time) {
	strcat(headers, "\r\nDate: ");
	strcat(headers, http_time(&cur_time));
    }
#endif	

    if (HTLocation) {		/* Do redirection */
	strcat(headers, "\r\nLocation: ");
	strcat(headers, HTLocation);
    }

    if (req->content_language) {
	strcat(headers, "\r\nContent-Language: ");
	strcat(headers, HTAtom_name(req->content_language));
    }

    if (req->content_encoding) {
	char * name = HTAtom_name(req->content_encoding);
	if (name && strcmp(name,"7bit") && strcmp(name,"8bit") &&
	    strcmp(name,"binary")) {
	    strcat(headers, "\r\nContent-Encoding: ");
	    strcat(headers, name);
	}
    }

    if (req->content_type) {
	strcat(headers, "\r\nContent-Type: ");
	strcat(headers, HTAtom_name(req->content_type));
    }

    if (req->content_length > 0) {
	char content_len[30];
	sprintf(content_len, "\r\nContent-Length: %d", req->content_length);
	strcat(headers, content_len);
    }

    if (HTLastModified) {
	strcat(headers, "\r\nLast-Modified: ");
	strcat(headers, HTLastModified);
    }

    if (HTExpires) {
	strcat(headers, "\r\nExpires: ");
	strcat(headers, HTExpires);
    }

    if (HTUri) {
	strcat(headers, "\r\nUri: ");
	strcat(headers, HTUri);
    }

    if (HTMessageId) {
	strcat(headers, "\r\nMessage-Id: ");
	strcat(headers, HTMessageId);
    }

    if (!HTList_isEmpty(HTPublic)) {
	HTList * cur = HTPublic;
	HTAtom * a;
	strcat(headers, "\r\nPublic:");
	while (NULL != (a = (HTAtom*)HTList_nextObject(cur))) {
	    strcat(headers, " ");
	    strcat(headers, HTAtom_name(a));
	}
    }

    if (!HTList_isEmpty(HTAllowed) || !HTList_isEmpty(HTPublic)) {
	HTList * cur;
	HTAtom * a;

	strcat(headers, "\r\nAllowed:");

	cur = HTPublic;
	while (NULL != (a = (HTAtom*)HTList_nextObject(cur))) {
	    strcat(headers, " ");
	    strcat(headers, HTAtom_name(a));
	}

	cur = HTAllowed;
	while (NULL != (a = (HTAtom*)HTList_nextObject(cur))) {
	    strcat(headers, " ");
	    strcat(headers, HTAtom_name(a));
	}
    }

    if (extras) {
	int i;
	for (i=0; i<extra_cnt; i++) {
	    if (extras[i] && *(extras[i])) {
		strcat(headers, "\r\n");
		strcat(headers, extras[i]);
	    }
	}
    }

    strcat(headers, "\r\n");
    if (HTWWWAuthenticate) {
	strcat(headers, HTWWWAuthenticate);
    }

    if (meta_file) {
	char * buf = (char*)malloc(1001);
	if (!buf) outofmem(__FILE__, "HTReplyHeaders");

	while (fgets(buf, 1000, meta_file)) {
	    if (fix_meta_field(buf))
		strcat(headers, buf);
	}
	fclose(meta_file);
	free(buf);		/* Leak fixed AL 27 Feb 1994 */
    }

    strcat(headers, "\r\n");

    VTRACE(stderr,
	   "............ Header section for client\n%s%s",
	   headers, "............ End of headers\n");

    return headers;
}



PUBLIC int HTLoadRedirection ARGS1(HTRequest *, req)
{
    if (req && HTLocation && HTReason == HTAA_OK_REDIRECT &&
	req->output_stream) {

	char * body = NULL;

	if (HTLocation[0] == '/') {
	    /* Need to fill in missing parts in the URL */
	    CONST char * host =
		HTServerHostName ? HTServerHostName : HTHostName();
	    char portstr[10];
	    char * new_loc = NULL;

	    if (HTServerPort > 0  &&  HTServerPort != 80)
		sprintf(portstr, ":%d", HTServerPort);
	    else
		*portstr = 0;

	    if (!host) host = "CANT.GET.MY.HOST.NAME";

	    new_loc = (char*)malloc(strlen(host)+strlen(HTLocation)+50);
	    if (!new_loc) outofmem(__FILE__, "HTLoadRedirection");
	    sprintf(new_loc, "http://%s%s%s", host, portstr, HTLocation);
	    free(HTLocation);
	    HTLocation = new_loc;
	}

	body = (char*)malloc(strlen(HTLocation) + 400);
	if (!body) outofmem(__FILE__, "HTLoadRedirection");
	sprintf(body, "%s%s%s%s%s\n",
		"<HEAD>\n<TITLE>Redirection</TITLE>\n</HEAD>\n<BODY>\n",
		"<H1>Redirection</H1>\nThis document can be found <A HREF=\"",
		HTLocation,
		"\">elsewhere.</A> <P>\nYou see this message because your ",
		"browser doesn't handle\nredirections automatically. </BODY>");
	req->content_type = HTAtom_for("text/html");
	req->content_length = strlen(body);

	if (HTClientProtocol) {
	    char * headers = HTReplyHeaders(req);
	    if (headers) {
		HTLoadStrToStream(req->output_stream, headers);
		free(headers);
	    }
	}
	if (req->method != METHOD_HEAD)
	    HTLoadStrToStream(req->output_stream, body);
	free(body);
	HTCloseStream(req->output_stream);
	return HT_LOADED;
    }
    return HT_INTERNAL;
}





/*	Converter for writing the MIME wrapper to a document
**	----------------------------------------------------
**
**	Thsi kinda cheats because it returns the sink stream
**	having first sent the wrapper down it.  Efficient -- I like it
**	better than the MIME parser which gets in the way from then on.
*/
PUBLIC HTStream * HTMIMEWrapper ARGS5(HTRequest *,	req,
				      void *,		param,
				      HTFormat,		input_format,
				      HTFormat,		output_format,
				      HTStream *,	sink)
{
    CTRACE(stderr, "Ok.......... Content-type %s\n",
	   HTAtom_name(input_format));

    req->content_type = input_format;

    if (HTClientProtocol) {
	char *headers = HTReplyHeaders(req);
	if (headers) {
	    (*sink->isa->put_string)(sink, headers);
	    free(headers);
	}
    } else {			/* Old protocol */
        if (input_format == WWW_PLAINTEXT) {
	    (*sink->isa->put_string)(sink, "<PLAINTEXT>\r\n");
	}
    }

    return sink;
}


PUBLIC HTStream * HTHackWrapper ARGS5(HTRequest *,	req,
				      void *,		param,
				      HTFormat,		input_format,
				      HTFormat,		output_format,
				      HTStream *,	sink)
{
    CTRACE(stderr,
    "Hack........ Sending to client even there was no corresponding Accept\n");
    CTRACE(stderr, "Ok.......... Content-type %s\n",
	   HTAtom_name(input_format));

    req->content_type = input_format;

    if (HTClientProtocol) {
	char * extras[2];
	char * headers;
	extras[0] =
    "X-Httpd-Warning: Your broser didn't send the Accept header line for this";
	extras[1] = NULL;
	headers = HTReplyHeadersWith(req, extras, 1);
	if (headers) {
	    (*sink->isa->put_string)(sink, headers);
	    free(headers);
	}
    } else {			/* Old protocol */
        if (input_format == WWW_PLAINTEXT) {
	    (*sink->isa->put_string)(sink, "<PLAINTEXT>\r\n");
	}
    }

    return sink;
}
    


PRIVATE char * get_part ARGS2(char **,	source,
			      char *,	separator)
{
    char * word;
    char * p;
    char * q;

    if (!source || !*source) return NULL;
    word = *source;

    while (*word && (*word==' ' || *word=='\t'))
	word++;

    for (p=word;  *p && *p != ';' && *p != ',';  p++) {
	if (*p == ' ') {	/* @@@ HACK TO UNDERSTAND OLD Accept: FORMAT */
	    *p = 0;		/* @@@ REMOVE THESE 4 LINES ONCE OLD         */
	    break;		/* @@@ BROWSERS DISAPPEAR (besides, this     */
	}			/* @@@ is VERY WRONG).			     */
    }
    *separator = *p;
    if (*p) {
	*source = p+1;
	*p = 0;
    }
    else *source = p;

    for (q=p-1; q>word; q--) {
	if (*q==' ' || *q=='\t')
	    *q = 0;
	else break;
    }

    return word;
}


PRIVATE char * get_word ARGS1(char **, source)
{
    char * word;
    char * p;

    if (!source || !*source) return NULL;
    word = *source;

    while (*word && (*word==' ' || *word=='\t'))
	word++;

    for (p=word;  *p && *p != ' ' && *p != '\t';  p++);
    while (*p && (*p == ' ' || *p == '\t'))
	*p++ = 0;
    *source = p;

    return word;
}



PRIVATE char * parse_entry ARGS5(char **,	source,
				 char **,	accept_x,
				 float *,	q,
				 float *,	mxs,
				 float *,	mxb)
{
    char sep;

    if (!source || !*source || !accept_x || !q || !mxs || !mxb)
	return NULL;

    *q = 1.0;
    *mxb = 0.0;
    *mxs = 0.0;
    *accept_x = get_part(source, &sep);
    if (!*accept_x || !**accept_x)
	return NULL;
    while (sep == ';') {
	float value;
	char *part = get_part(source, &sep);
	char *equals = part ? strchr(part, '=') : NULL;

	if (!equals) break;    /* bad syntax -- forget it! */
	*equals++ = 0;	/* Split at equals */
	if (sscanf(equals, "%f", &value)) {
	    char * attrib = HTStrip(part);
	    if (!strcasecomp(attrib, "q"))
		*q = value;
	    else if (!strcasecomp(attrib, "mxb"))
		*mxb = value;
	    else if (!strcasecomp(attrib, "mxs"))
		*mxs = value;
	}
    } /* scan attributes */
    return *accept_x;
}



/* PUBLIC						HTParseRequest()
**		READ REQUEST FROM HTInputSocket OBJECT
**		AND PARSE IT INTO A HTRequest STRUCTURE
** ON ENTRY:
**	isoc	input socket structure.
**
** ON EXIT:
**	returns	a newly allocated HTRequest object describing
**		parsed request.
**		NULL if something fails badly.
**	isoc	is read until the end of header section (empty
**		line is last thing read), so isoc input pointer
**		points to the beginning of message body.
*/
PUBLIC HTRequest * HTParseRequest ARGS1(HTInputSocket *, isoc)
{
    HTRequest *req;
    char *line;
    char *p, *p1, *p2, *p3;
    BOOL text_html = NO;
    BOOL text_plain = NO;

    if (!isoc)
	return NULL;

    req = HTRequest_new();
    HTReason = HTAA_OK;
    HTClientProtocol = NULL;	/* Assume HTTP0 */

    /*
    ** Parse first line of request
    */
    p = line = HTInputSocket_getLine(isoc);
    if (line) {
	VTRACE(stderr, "Client sez.. %s\n", line);
	StrAllocCopy(HTReqLine, line);

	p1 = get_word(&p);
	p2 = get_word(&p);
	p3 = get_word(&p);

	if (p1) req->method = HTMethod_enum(p1);
	if (p2) StrAllocCopy(HTReqArg, p2);
	if (p3) StrAllocCopy(HTClientProtocol, p3);
	free(line);	/* Leak fixed AL 6 Feb 94 */
    }

    /*
    ** Check that the third argument actually is a valid
    ** client protocol specifier (if it is not we might wait
    ** for an eternity for the rest of an HTTP1 request when it
    ** will never come, becuase it's actually an HTTP0 request.
    */
    if (HTClientProtocol &&
	0 != strncasecomp(HTClientProtocol, "HTTP/", 5))
	HTClientProtocol = NULL;

#ifndef NOCONVERT
    if (!req->conversions)
	req->conversions = HTList_new();
    if (do_accept_hack)
	HTSetConversion(req->conversions,
			"*/*", "www/present", HTHackWrapper,
			0.00001, 0.0, 0.0);
#endif			
    
    hbuf_free();	/* From previous call */

    if (HTClientProtocol) {
      
	while ((line = HTInputSocket_getUnfoldedLine(isoc))) {

	    hbuf_append(line);

	    if (!*line) {
		free(line);	/* Leak fixed AL 6 Feb 1994 */
		break;		/* Empty line terminating header section */
	    }
	    VTRACE(stderr, "Client sez.. %s\n", line);
	    p = strchr(line, ':');
	    if (!p) continue;			/* Bad format -- junk line */
	    *p++ = 0;					/* Overwrite colon */
	    for ( ;  *p==' ' || *p=='\t';  p++);	/* Skip whitespace */

	    if (!strcasecomp(line, "Accept") ||
		!strcasecomp(line, "Accept-Encoding") ||
		!strcasecomp(line, "Accept-Language")) {
		char *name = NULL;
		float quality;
		float maxbytes;
		float maxsecs;

		while (parse_entry(&p,&name,&quality,&maxbytes,&maxsecs)) {  
		    if (!strcasecomp(line, "Accept")) {

#ifdef VMS	/* In Unix we do this in the general HTTP_XXX way */
			hbuf_cgi_cancel_last();
#endif
			VTRACE(stderr,
			       "Accept...... %s (q=%.2f,mxb=%.1f,mxs=%.1f)\n",
			       name, quality, maxbytes, maxsecs);
#ifndef NOCONVERT
			if (!strcasecomp(name, "text/html"))
			    text_html = YES;
			else if (!strcasecomp(name, "text/plain"))
			    text_plain = YES;

			if (!strcmp(name, "www/source"))
			    HTSetConversion(req->conversions,
					    "*/*","www/present",HTMIMEWrapper,
					    quality, maxbytes, maxsecs);
			else
			    HTSetConversion(req->conversions,
					    name, "www/present", HTMIMEWrapper,
					    quality, maxbytes, maxsecs);
#endif
		    }
		    else if (!strcasecomp(line, "Accept-Encoding")) {
			CTRACE(stderr, "Encoding.... %s (q=%.2f)\n",
			       name, quality);
			if (!req->encodings)
			    req->encodings = HTList_new();
			HTAcceptEncoding(req->encodings, name, quality);
		    }
		    else {
			CTRACE(stderr, "Language.... %s (q=%.2f)\n",
			       name, quality);
			if (!req->languages)
			    req->languages = HTList_new();
			HTAcceptLanguage(req->languages, name, quality);
		    }
		}

	    } else if (0==strncasecomp(line, "Content-Type", 12)) {

		hbuf_cgi_cancel_last();
		p1 = get_word(&p);
		if (p1) {
		    req->content_type = HTAtom_for(p1);
		    CTRACE(stderr, "Content-Type %s\n", p1);
		}

	    } else if (0==strncasecomp(line, "Content-Length", 14)) {

		hbuf_cgi_cancel_last();
		p1 = get_word(&p);
		if (p1) {
		    req->content_length = atoi(p1);
		    CTRACE(stderr, "Content-Length %d\n", req->content_length);
		}

	    } else if (0==strncasecomp(line, "Authorization", 13)) {

		StrAllocCopy(req->authorization, p);
		p1 = get_word(&p);
		p2 = get_word(&p);
		CTRACE(stderr, "Authorization scheme \"%s\" params \"%s\"\n",
		       (p1 ? p1 : "-null-"),  (p2 ? p2 : "-null-"));
		if (p1 && p2) {
		    req->scheme = HTAAScheme_enum(p1);
		    StrAllocCopy(HTAuthString, p2);
		}
		else CTRACE(stderr, "Invalid..... Authorization: field\n");

	    } else if (0==strncasecomp(line, "From", 4)) {

		StrAllocCopy(req->from, HTStrip(p));
		CTRACE(stderr, "From........ %s\n", req->from);

	    } else if (0==strncasecomp(line, "User-Agent", 10)) {

		hbuf_proxy_cancel_last();
		StrAllocCopy(HTUserAgent, HTStrip(p));
		CTRACE(stderr, "User-Agent.. %s\n", HTUserAgent);

	    } else if (0==strncasecomp(line, "Referer", 7)) {

		StrAllocCopy(HTReferer, HTStrip(p));
		CTRACE(stderr, "Referer..... %s\n", HTReferer);

	    } else if (0==strncasecomp(line, "If-Modified-Since", 17)) {

		HTIfModifiedSince = parse_http_time(HTStrip(p));
		CTRACE(stderr, "Give only... if modified since (localtime) %s",
		       ctime(&HTIfModifiedSince));

	    } else if (0==strncasecomp(line, "Charge-To", 9)) {

		/* Add code here */

	    }
	    /* else invalid */
	    free(line);		/* Leak fixed AL 6 Feb 94 */

	} /* scan lines */
    } /* if HTTP1 protocol */
#ifndef NOCONVERT
    else {
	/*
	** Old protocol browsers have to be assumed to allow binary and
	** plain old source.  This is for XMosaic 1.x, and we don't have
	** to support it.. soon.
	** This is risky & more trouble.
	*/	
	HTSetConversion(req->conversions,
			"application/octet-stream","www/present",HTMIMEWrapper,
			0.7, 0.0, 0.0);
	HTSetConversion(req->conversions,
			"www/source", "www/present", HTMIMEWrapper,
			0.7, 0.0, 0.0);
    }
#endif

#ifndef NOCONVERT
    if (!text_html)
	HTSetConversion(req->conversions,
			"text/html", "www/present", HTMIMEWrapper,
			1.0, 0.0, 0.0);
    if (!text_plain)
	HTSetConversion(req->conversions,
			"text/plain", "www/present", HTMIMEWrapper,
			1.0, 0.0, 0.0);
#endif			

    compute_server_env();

    if (HTImProxy) {
	HTProxyHeaders = hbuf_proxy_headers(req);
    }

    return req;
}

